/************************************************************************/
/*      This program reformats the binary output of the Magellan AIV-10 */
/*      OEM GPS receiver into ASCII files for further data processing.  */
/*      The program takes as input a file containing the raw binary     */
/*      data from the OEM board it produces two files as output.        */
/*      One, with the extension .OUT, contains one record per message   */
/*      set from the OEM board with most of the information transformed */
/*      into ASCII.  The other, with the extension .PVT, contains one   */
/*      record per message set from the OEM board with only fix time    */
/*      position and velocity.                                          */
/*                                                                      */
/*      NOTE:  The program assumes that output messages type 81, 82 and */
/*      83 (hex) are all enabled at the same rate.  If this is not the  */
/*      case the logic in this program will need to be modified.        */
/*                                                                      */
/*      The program can be executed with the input file's name on the   */
/*      command line (DEC10 file) or simply by typing the program name  */
/*      (DEC10).  In the later case the program will prompt for the     */
/*      input file's name.  If no extension is provided in the input    */
/*      file's name and the base name file cannot be opened the         */
/*      extension .LOG will automatically be appended to the given      */
/*      file name.  For example:                                        */
/*              DEC10 RAW.BIN                                           */
/*      will only try to open the file RAW.BIN for input.               */
/*              DEC10 RAW                                               */
/*      will try to open RAW. for input and if this fails will then     */
/*      try to open RAW.LOG.  In either case a path name may be pre-    */
/*      pended to the input file's name.                                */
/*                                                                      */
/*      Output file names are constructed from the base name of the     */
/*      input file with the extensions .OUT and .PVT.  For instance,    */
/*      in the examples above the output files would be RAW.OUT and     */
/*      RAW.PVT.  Output files are written to the same directory as     */
/*      contains the input file.                                        */
/*                                                                      */
/*                                                                      */
/*      OUTPUT FILE FORMATS:                                            */
/*                                                                      */
/*      Records in the .PVT file:                                       */
/*         Fix time in seconds into the week.  8 columns                */
/*         1 blank column                                               */
/*         Position in 26 columns:                                      */
/*           If OEM output is in latitude and longitude:                */
/*              Latitude in degrees - 9 columns                         */
/*              1 blank column                                          */
/*              Longitude in degrees - 10 columns                       */
/*              1 blank column                                          */
/*              Altitude (in feet or meters as specified by OEM output  */
/*                        format) - 5 columns                           */
/*           If OEM output is in UTM:                                   */
/*              Northing in meters - 8 columns                          */
/*              1 blank column                                          */
/*              Easting in meters - 8 columns                           */
/*              1 blank column                                          */
/*              Altitude in meters - 8 columns                          */
/*           If OEM output is in ECEF:                                  */
/*              ECEF-X in meters - 8 columns                            */
/*              1 blank column                                          */
/*              ECEF-Y in meters - 8 columns                            */
/*              1 blank column                                          */
/*              ECEF-Z in meters - 8 columns                            */
/*         1 blank column                                               */
/*         Velocity in 14 columns (units are as specified by OEM        */
/*                      output):                                        */
/*           If OEM output is in North/East/Down:                       */
/*              North velocity - 4 columns                              */
/*              1 blank column                                          */
/*              East velocity - 4 columns                               */
/*              1 blank column                                          */
/*              Down velocity - 4 columns                               */
/*           If OEM output is in speed and heading:                     */
/*              Speed - 8 columns                                       */
/*              1 blank column                                          */
/*              Heading in degrees - 5 columns                          */
/*                                                                      */
/*      Records in the .OUT file:                                       */
/*         UTC time in seconds into day - 5 columns                     */
/*         1 blank column                                               */
/*         Fix time in seconds into the week.  8 columns                */
/*         1 blank column                                               */
/*         Position in 26 columns:                                      */
/*           If OEM output is in latitude and longitude:                */
/*              Latitude in degrees - 9 columns                         */
/*              1 blank column                                          */
/*              Longitude in degrees - 10 columns                       */
/*              1 blank column                                          */
/*              Altitude (in feet or meters as specified by OEM output  */
/*                        format) - 5 columns                           */
/*           If OEM output is in UTM:                                   */
/*              Northing in meters - 8 columns                          */
/*              1 blank column                                          */
/*              Easting in meters - 8 columns                           */
/*              1 blank column                                          */
/*              Altitude in meters - 8 columns                          */
/*           If OEM output is in ECEF:                                  */
/*              ECEF-X in meters - 8 columns                            */
/*              1 blank column                                          */
/*              ECEF-Y in meters - 8 columns                            */
/*              1 blank column                                          */
/*              ECEF-Z in meters - 8 columns                            */
/*         1 blank column                                               */
/*         Velocity in 14 columns (units are as specified by OEM        */
/*                      output):                                        */
/*           If OEM output is in North/East/Down:                       */
/*              North velocity - 4 columns                              */
/*              1 blank column                                          */
/*              East velocity - 4 columns                               */
/*              1 blank column                                          */
/*              Down velocity - 4 columns                               */
/*           If OEM output is in speed and heading:                     */
/*              Speed - 8 columns                                       */
/*              1 blank column                                          */
/*              Heading in degrees - 5 columns                          */
/*         Ten sets of receiver channel status:                         */
/*           1 blank column                                             */
/*           2 digit hex status code for the channel                    */
/*           Slash character ('/')                                      */
/*           2 digit PRN of the satellite assigned to the channel       */
/*           Slash character ('/')                                      */
/*           2 digit value of reported SNR divided by 3                 */
/*         1 blank column                                               */
/*         Character '1' if fix was updated, '0' if not                 */
/*         Slash character ('/')                                        */
/*         Character '1' if fix was from dead-reckoning, '0' if not     */
/*         Slash character ('/')                                        */
/*         2 digit count of the number of satellites used in the fix    */
/*         Slash character ('/')                                        */
/*         2 digit count of the number of satellites currently in track */
/*         For each satellite used in the position fix:                 */
/*           Slash character ('/')                                      */
/*           2 digit PRN of the satellite                               */
/*         1 blank column                                               */
/*         2 digit PDOP value (if >99, '99' is output)                  */
/*         Slash character ('/')                                        */
/*         2 digit HDOP value (if >99, '99' is output)                  */
/*         Slash character ('/')                                        */
/*         2 digit VDOP value (if >99, '99' is output)                  */
/*         Slash character ('/')                                        */
/*         3 digit error estimate value in meters (if >999, '999' is    */
/*                      output)                                         */
/*                                                                      */
/*      This program was debugged and compiled under Microsoft C v6.00A */
/************************************************************************/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>


/************************************************************************/
/*      This routine swaps the bytes of a long integer (4 bytes) to     */
/*      convert from Motorola format (as provided by OEM board) to      */
/*      Intel format.                                                   */

static long int byte4(unsigned char *loc_buf) {
  union {
     long int           lint;
     unsigned char      byte[4];
  } swapper;

    swapper.byte[3] = *loc_buf++;
    swapper.byte[2] = *loc_buf++;
    swapper.byte[1] = *loc_buf++;
    swapper.byte[0] = *loc_buf;
    return swapper.lint;
}


/************************************************************************/
/*      This routine swaps the bytes of a short integer (2 bytes) to    */
/*      convert from Motorola format (as provided by OEM board) to      */
/*      Intel format.                                                   */

static short int byte2(unsigned char *loc_buf) {
  union {
     short int          word;
     unsigned char      byte[2];
  } swapper;

    swapper.byte[1] = *loc_buf++;
    swapper.byte[0] = *loc_buf;
    return swapper.word;
}


/************************************************************************/
/************************   MAIN PROGRAM START   ************************/
/************************************************************************/

static unsigned char buffer[2048];      /* Raw data input buffer        */


void main(int argc, char *argv[]) {
  char          file[128];      /* Buffer for file names                */
  char          *chr_ptr;       /* Used to parse file names             */
  unsigned int  length;         /* Raw message length                   */
  int           i, j;           /* Loop counter/temporary               */
  unsigned char csum;           /* Raw message checksum accumulator     */
  long int      pos_time;       /* Fix time in miliseconds              */
  long int      rep_time;       /* Current time from OEM                */
  long int      pos[3];         /* Position                             */
  int           num_pos_sats;   /* Number of sats in position fix       */
  int           vel[3];         /* Velocity                             */
  unsigned char pos_sats[10];   /* PRNs of sats in position fix         */
  unsigned int  prn[10];        /* PRNs of sats tracked                 */
  unsigned int  snr[10];        /* SNRs of sats tracked                 */
  unsigned int  errest;         /* Estimated position error (meters)    */
  FILE          *input;         /* Input file                           */
  FILE          *output;        /* Primary (.OUT) output file           */
  FILE          *out2;          /* Secondary (.PVT) output file         */
  float         pdop, hdop, vdop;/* DOPs of position fix                */
  char          dr_pos;         /* Set true if fix was from dead-       */
                                /*      reckoning                       */
  char          pos_type;       /* Type of position reported (bits 5 &  */
                                /*      6 of subtype of message 81)     */
  char          vel_type;       /* Type of velocity reported (bit 4 of  */
                                /*      subtype of message 81)          */
  char          new_pos = 0;    /* Flag set true if current fix is new  */
  int           num_tracked = 0;/* Number of satellites tracked         */
  int           got_all = 0;    /* Bit flags for messages received      */
                                /*      (bit 0 for type 81, bit 1 for   */
                                /*       type 82)                       */
  int           status[10] =    /* Status of tracked satellites         */
                        {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
  long int      prev_pos_time = 0;      /* Time of previous fix         */


/* See if there was a file specified on command line, if not get one    */
/* from user                                                            */

    if (argc == 1) {
       printf("Enter raw data file: ");
       (void)gets(file);
    } else {
       strcpy(file,argv[1]);
    }

/* Attempt to open input file.  If necessary append .LOG extension.     */
/* If we can't open the input file terminate the program.               */

    input = fopen(file,"rb");
    if (input == NULL) {
       for (i = (strlen(file)-1); i > 0; i--) {
          if (file[i] == '.') {
             printf("Could not open input file %s\007\n",file);
             exit(1);
          } else if (file[i] == '\\') {
             break;
          }
       }
       strcat(file,".LOG");
       input = fopen(file,"rb");
       if (input == NULL) {
          printf("Could not open input file %s\007\n",file);
          exit(1);
       }
    }

/* Construct output file names and open them.  If they don't open       */
/* terminate the program.                                               */

    for (chr_ptr = &(file[strlen(file)-1]);
         (chr_ptr != file) && (*chr_ptr != '.') && (*chr_ptr != '\\');
         chr_ptr--) ;
    if ((*chr_ptr == '\\') || (chr_ptr == file)) {
       chr_ptr = &(file[strlen(file)-1]);
       strcat(file,".OUT");
    } else {
       strcpy(chr_ptr,".OUT");
    }
    output = fopen(file,"w");
    if (output == NULL) {
       printf("Could not open output file %s\007\n",file);
       exit(1);
    }
    strcpy(chr_ptr,".PVT");
    out2 = fopen(file,"w");
    if (out2 == NULL) {
       printf("Could not open output file %s\007\n",file);
       exit(1);
    }

/* Output header line to primary output file    */

    fprintf(output," REP     POS\n");
    fprintf(output," tod     tow               POS                   VEL                                10 channels of status/prn/(snr/3)                              f/d/fs/ns fix-sats pdop/hdop/vdop/err\n");
    fprintf(output,"----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n");

/* This loop is continuous until the input file is exhausted of data    */

    do {

/* Get data from the input file until we identify a message header or   */
/* we hit the end of the file                                           */

       do {
          buffer[0] = fgetc(input);
          if (buffer[0] == '$') {
             buffer[1] = fgetc(input);
          }
       } while (!feof(input) &&
                ((buffer[0] != '$') || (buffer[1] != '$')));

       if (!feof(input)) {

/* Read the message length from the header and then read the rest of    */
/* the message                                                          */

          buffer[2] = fgetc(input);
          buffer[3] = fgetc(input);
          length = 256*buffer[2] + buffer[3];

                        /* Only process valid length messages   */
          if ((length == 46) || (length == 12) || (length == 70) ||
              (length == 1028) || (length == 20)) {
             (void)fread(&(buffer[4]),1,length,input);

/* Check for the end of message character and verify the checksum       */

             if ((buffer[4+length-1] == 10) && !feof(input)) {
                  /* Check checksum       */
                csum = 0;
                j = length + 2;
                for (i = 0; i < j; i++) {
                   csum = csum ^ buffer[i];
                }
                if (csum == buffer[4+length-2]) {

/* At this point we have a complete message which has a valid checksum  */
/* Now we process the different types of messages.                      */

                   switch (buffer[4]) {

/* POSITION */
                     case 0x081:
                       pos_type = buffer[5] & 0x060;
                       vel_type = buffer[5] & 0x010;
                       dr_pos = (buffer[6] & 0x004) >> 2;
                       pos_time = byte4(&(buffer[7]));

                        /* Base position updated flag on time difference*/
                       new_pos = (pos_time != prev_pos_time);
                       prev_pos_time = pos_time;
                       pos[0] = byte4(&(buffer[11]));
                       pos[1] = byte4(&(buffer[15]));
                       if (pos_type == 0x060) {
                          pos[2] = byte2(&(buffer[19]));
                       } else {
                          pos[2] = byte4(&(buffer[19]));
                       }
                       vel[0] = byte2(&(buffer[23]));
                       vel[1] = byte2(&(buffer[25]));
                       vel[2] = byte2(&(buffer[27]));
                       pdop = 0.01*(unsigned int)byte2(&(buffer[29]));
                       hdop = 0.01*(unsigned int)byte2(&(buffer[31]));
                       vdop = 0.01*(unsigned int)byte2(&(buffer[33]));
                       errest = (unsigned int)byte2(&(buffer[35]));
                       num_pos_sats = buffer[37];
                       memcpy(pos_sats,&(buffer[38]),10);
                       got_all = got_all | 1;
                       break;

/* TIME */
                     case 0x082:
                       rep_time = 3600L*(unsigned long int)buffer[10] +
                                  60L*(unsigned long int)buffer[11] +
                                  buffer[12];
                       got_all = got_all | 2;
                       break;

/* RECEIVER STATUS      */
/* This is the last message from a given output cycle that is   */
/* processed.  So this is used to trigger output to the files   */
                     case 0x083:
                       if (got_all == 3) {

        /* Get status information       */

                          num_tracked = buffer[11];
                          for (i = 0; i < 10; i++) {
                             status[i] = buffer[12+6*i];
                             prn[i] = buffer[12+6*i+1];
                             snr[i] = buffer[12+6*i+2];
                          }

        /* Output time and position     */

                          fprintf(output,"%5ld %8.1lf",rep_time,
                                  ((double)pos_time/1000.0));
                          fprintf(out2,"%8.1lf",pos_time/1000.0);
                          switch (pos_type) {
                            case 0x000:
                            case 0x020:
                              fprintf(output,"%10.5lf %10.5lf %5ld",
                                      ((double)pos[0]*0.0000001),
                                      ((double)pos[1]*0.0000001),
                                      pos[2]);
                              fprintf(out2,"%10.5lf %10.5lf %5ld",
                                      ((double)pos[0]*0.0000001),
                                      ((double)pos[1]*0.0000001),
                                      pos[2]);
                              break;
                            case 0x040:
                            case 0x060:
                              fprintf(output," %8ld %8ld %8ld",
                                      pos[0],pos[1],pos[2]);
                              fprintf(out2," %8ld %8ld %8ld",
                                      pos[0],pos[1],pos[2]);
                              break;
                          }

        /* Output velocity      */

                          if (vel_type) {
                             fprintf(output," %4d %4d %4d",
                                     (vel[0]/10),(vel[1]/10),
                                     (vel[2]/10));
                             fprintf(out2," %4d %4d %4d\n",
                                     (vel[0]/10),(vel[1]/10),
                                     (vel[2]/10));
                          } else {
                             fprintf(output," %8.1lf %5.1lf",
                                     ((double)vel[0]*0.1),
                                     ((unsigned int)vel[1]*0.01));
                             fprintf(out2," %8.1lf %5.1lf\n",
                                     ((double)vel[0]*0.1),
                                     ((unsigned int)vel[1]*0.01));
                          }

        /* Output receiver status       */

                          for (i = 0; i < 10; i++) {
                             if (status[i] == 0) {
                                fprintf(output," 00/00/00");
                             } else {
                                fprintf(output," %02x/%02d/%02d",
                                        status[i],prn[i],(snr[i]/3));
                             }
                          }

        /* Output fix flags, fix sats and DOPs  */

                          fprintf(output," %1d/%1d/%02d/%02d",new_pos,
                                  dr_pos,num_pos_sats,num_tracked);
                          for (i = 0; i < num_pos_sats; i++) {
                             fprintf(output,"/%02d",(int)pos_sats[i]);
                          }
                          i = pdop + 0.5;
                          if (i < 100) {
                             fprintf(output," %02d/",i);
                          } else {
                             fprintf(output," 99/");
                          }
                          i = hdop + 0.5;
                          if (i < 100) {
                             fprintf(output,"%02d/",i);
                          } else {
                             fprintf(output,"99/");
                          }
                          i = vdop + 0.5;
                          if (i < 100) {
                             fprintf(output,"%02d/",i);
                          } else {
                             fprintf(output,"99/");
                          }
                          if (errest < 1000) {
                             fprintf(output,"%03d\n",errest);
                          } else {
                             fprintf(output,"999\n");
                          }
                       }
                       got_all = 0;
                       break;

/* GENERAL STATUS - ignored     */
                     case 0x084:
                       break;

/* SETUP - ignored      */
                     case 0x085:
                       break;

/* ALMANAC - ignored    */
                     case 0x086:
                       break;
                   }
                }
             }
          }
       }
    } while (!feof(input));

/* Successful completion, close files and exit  */

    fclose(output);
    fclose(out2);
    fclose(input);
    exit(0);
}
